/* Copyright (C) 2003-2005 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
#include <idna.h>
#endif

#include "dps_common.h"
#include "dps_socket.h"
#include "dps_host.h"
#include "dps_proto.h"
#include "dps_utils.h"
#include "dps_mutex.h"
#include "dps_log.h"


#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

#define MAX_HOST_ADDR_SIZE	4096


#define DEBUG 1


static int cmphost(const void * v1, const void * v2){
	return(strcasecmp(((const DPS_HOST_ADDR*)v1)->hostname, ((const DPS_HOST_ADDR*)v2)->hostname));
}


static int host_addr_add(DPS_HOSTLIST *List, const char *hostname, struct in_addr *addr) {
	int min_id;
	
	if((List->nhost_addr>=List->mhost_addr)&&(List->mhost_addr<MAX_HOST_ADDR_SIZE)){
		List->mhost_addr += MAX_HOST_ADDR_SIZE; /* FIXME: was 32 */
		List->host_addr = DpsRealloc(List->host_addr, List->mhost_addr*sizeof(DPS_HOST_ADDR));
		if (List->host_addr == NULL) {
		  List->mhost_addr = List->nhost_addr = 0;
		  return DPS_ERROR;
		}
		bzero((void*)&List->host_addr[List->nhost_addr], (List->mhost_addr-List->nhost_addr)*sizeof(DPS_HOST_ADDR));
	}

	if((List->nhost_addr<List->mhost_addr)&&(List->mhost_addr<=MAX_HOST_ADDR_SIZE)){
		min_id = List->nhost_addr;
		List->nhost_addr++;
	}else{
		int min;
		size_t i;

		min = List->host_addr[0].last_used;
		min_id=0;
		/* find last used host */
		for (i=0; i<List->nhost_addr; i++)
			if(List->host_addr[i].last_used < List->host_addr[min_id].last_used){
				min = List->host_addr[i].last_used;
				min_id=i;
			}
	}

	List->host_addr[min_id].last_used = time(NULL);
	if(addr)memcpy(&List->host_addr[min_id].addr, addr, sizeof( struct in_addr));
	
	DPS_FREE(List->host_addr[min_id].hostname);
	List->host_addr[min_id].hostname = (char*)DpsStrdup(hostname);
	List->host_addr[min_id].net_errors=0;

	DpsSort(List->host_addr, List->nhost_addr, sizeof(DPS_HOST_ADDR), cmphost);
	return 0;
}

static DPS_HOST_ADDR *host_addr_find(DPS_HOSTLIST *List, const char *hostname) {
/*	int l,c,r,res;*/
        DPS_HOST_ADDR key;

        /* Find current URL in sorted part of list */

	if (List->nhost_addr == 0) return NULL;
	key.hostname = hostname;
	return bsearch(&key, List->host_addr, List->nhost_addr, sizeof(DPS_HOST_ADDR), cmphost);

/*	

        l=0;r=List->nhost_addr-1;
        while(l<=r){
                c=(l+r)/2;
		if(!(res = strcasecmp(List->host_addr[c].hostname, hostname)))
                        return(&List->host_addr[c]);
                
                if(res<0)
                        l=c+1;
                else
                        r=c-1;
        }
	return NULL;
*/
}


int DpsHostLookup(DPS_AGENT *Indexer, DPS_CONN *connp) {
	DPS_HOST_ADDR  *Host;
	DPS_HOSTLIST *List = &Indexer->Hosts;
	char   *ascii = NULL;
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
	DPS_CHARSET *url_cs, *uni_cs;
	DPS_CONV  url_uni;
	char    *uni = NULL;
	size_t len;
	int rc;
#endif
	TRACE_IN(Indexer, "DpsHostLookup");

	if (!connp->hostname) {
	        TRACE_OUT(Indexer);
		return -1;
	}
	
#if defined(WITH_TRACE) && defined(DEBUG)
	fprintf(Indexer->TR, "Resolver looking for %s\n", connp->hostname);
	fflush(Indexer->TR);
#endif

	bzero((void*)&connp->sin, sizeof(struct sockaddr_in));
	
	/*	Check port	*/
	if (connp->port == 0) {
	  DpsLog(Indexer, DPS_LOG_DEBUG, "Zero port at %s:%d", __FILE__, __LINE__);
		connp->err = DPS_NET_ERROR;
		TRACE_OUT(Indexer);
		return -2;
	}

	/* Check if hostname in standard numbers-and-dots notation */ 
	if ((connp->sin.sin_addr.s_addr = inet_addr(connp->hostname)) == INADDR_NONE) {
	        int i;

		if((Host=host_addr_find(List,connp->hostname))){
			Host->last_used=time(NULL);
			connp->Host = Host;
			if(Host->addr.s_addr){
				memcpy(&connp->sin.sin_addr, &Host->addr, sizeof(struct in_addr));
			}else{
				connp->err = DPS_NET_CANT_RESOLVE;
				TRACE_OUT(Indexer);
				return -3;
			}
			/*	Set port	*/
			connp->sin.sin_port = htons(connp->port);
#if defined(WITH_TRACE) && defined(DEBUG)
	fprintf(Indexer->TR, "Resolver find: %s - > %s [Host.addr: %s]\n", connp->hostname, inet_ntoa(connp->sin.sin_addr),
		inet_ntoa(connp->Host->addr));
	fflush(Indexer->TR);
#endif
			TRACE_OUT(Indexer);
			return 0;
		}

#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
		if (connp->charset_id != DPS_CHARSET_US_ASCII) {
		  uni_cs = DpsGetCharSet("UTF-8");
		  url_cs = DpsGetCharSetByID(connp->charset_id);
		  DpsConvInit(&url_uni, url_cs, uni_cs, DPS_RECODE_URL);

		  uni = (char*)DpsMalloc(len = (48 * strlen(connp->hostname) + sizeof(dpsunicode_t)));
		  if (uni == NULL) {
		    connp->err = DPS_NET_CANT_RESOLVE;
		    TRACE_OUT(Indexer);
		    return -4;
		  }
		  DpsConv(&url_uni, (char*)uni, len, connp->hostname, len);
		  if ( (rc = idna_to_ascii_8z((const char*)uni, &ascii, 0)) != IDNA_SUCCESS) {
		    DpsLog(Indexer, DPS_LOG_ERROR, "IDN: %s [%s] -> %s [UTF-8]  rc:%d", connp->hostname, url_cs->name, (char*)uni, rc);
		    DPS_FREE(uni);
		    connp->err = DPS_NET_CANT_RESOLVE;
		    TRACE_OUT(Indexer);
		    return -5;
		  }
		  DpsLog(Indexer, DPS_LOG_DEBUG, "IDN: %s [%s] -> %s", connp->hostname, url_cs->name, ascii);
		}
#else
		ascii = connp->hostname;
#endif

#if defined(HAVE_PTHREAD) && defined(HAVE_GETADDRINFO)
		{
		  struct addrinfo hints, *res = NULL, *list;
		  int error = EAI_AGAIN;

		  memset(&hints, 0, sizeof(hints));
		  hints.ai_family = PF_INET;
		  hints.ai_socktype = SOCK_STREAM;

#if defined(__FreeBSD__) && !defined(HAVE_LIBBIND) && !defined(HAVE_LIBDNS) /* FreeBSD havn't thread safe getaddrinfo in libc */
		  DPS_GETLOCK(Indexer, DPS_LOCK_RESOLV);
#endif
#if defined(WITH_TRACE) && defined(DEBUG)
		  fprintf(Indexer->TR, "Resolver calling getaddrinfo\n");
		  fflush(Indexer->TR);
#endif
		  for (i = 0; (i < 5) && (error == EAI_AGAIN); i++) {
#if defined(WITH_TRACE) && defined(DEBUG)
		    fprintf(Indexer->TR, "Pas: %d begin\n", i);
		    fflush(Indexer->TR);
#endif
		    error = getaddrinfo(ascii, NULL, &hints, &res);
#if defined(WITH_TRACE) && defined(DEBUG)
		    fprintf(Indexer->TR, "Pas: %d done\n", i);
		    fflush(Indexer->TR);
#endif
		  }
		  if (error == EAI_NONAME) {
		    error = EAI_AGAIN;
		    DpsLog(Indexer, DPS_LOG_DEBUG, "%s not found, trying original %", ascii, connp->hostname);
		    for (i = 0; (i < 5) && (error == EAI_AGAIN); i++) {
#if defined(WITH_TRACE) && defined(DEBUG)
		      fprintf(Indexer->TR, "2Pas: %d begin\n", i);
		      fflush(Indexer->TR);
#endif
		      error = getaddrinfo(connp->hostname, NULL, &hints, &res);
#if defined(WITH_TRACE) && defined(DEBUG)
		      fprintf(Indexer->TR, "2Pas: %d done\n", i);
		      fflush(Indexer->TR);
#endif
		    }
		  }
		  if (error || res == NULL) {
		    if (error) {
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
		      DpsLog(Indexer, DPS_LOG_WARN, "IDN: %s [%s] -> %s", connp->hostname, url_cs->name, ascii);
#endif
		      DpsLog(Indexer, DPS_LOG_WARN, "%s: getaddrinfo error: %s", ascii, gai_strerror(error));
		      
		    } else {
		      DpsLog(Indexer, DPS_LOG_WARN, "%s: no adresses", ascii);
		    }
/*			host_addr_add(List,connp->hostname, NULL);*/
		    connp->err = DPS_NET_CANT_RESOLVE;
#if defined(__FreeBSD__) && !defined(HAVE_LIBBIND) && !defined(HAVE_LIBDNS) /* FreeBSD havn't thread safe getaddrinfo in libc */
		    DPS_RELEASELOCK(Indexer, DPS_LOCK_RESOLV);
#endif
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
		    DPS_FREE(uni); DPS_FREE(ascii);
#endif
		    TRACE_OUT(Indexer);
		    return -6;
		  }
#if defined(WITH_TRACE) && defined(DEBUG)
		  fprintf(Indexer->TR, "Resolver looking result\nconnp.timeout: %d\n", connp->timeout);
		  fflush(Indexer->TR);
#endif
		  for(list = res, i = 0; (list != NULL) && (i < 20); list = list->ai_next, i++) {
		    memcpy((char *)&(connp->sin), list->ai_addr, (size_t)list->ai_addrlen);
		    connp->sin.sin_port = htons(connp->port); /*	Set port	*/
		    DpsLog(Indexer, DPS_LOG_DEBUG, "Resolver %dth checking for %s", i, inet_ntoa(connp->sin.sin_addr));
		    if ((list->ai_next != NULL) && (DpsCheckAddr(&(connp->sin), connp->timeout) == DPS_OK)) break;
		  }
		  DpsLog(Indexer, DPS_LOG_DEBUG, "Resolver: %s - > %s", ascii, inet_ntoa(connp->sin.sin_addr));
		  if (res != NULL) freeaddrinfo(res);
#if defined(__FreeBSD__) && !defined(HAVE_LIBBIND) && !defined(HAVE_LIBDNS) /* FreeBSD havn't thread safe getaddrinfo in libc */
		  DPS_RELEASELOCK(Indexer, DPS_LOCK_RESOLV);
#endif
		}
#elif defined(HAVE_PTHREAD) && (defined(HAVE_GETHOSTBYNAME_R_3) || defined(HAVE_GETHOSTBYNAME_R_5) || defined(HAVE_GETHOSTBYNAME_R_6))
		{
		  struct hostent *he = NULL;
		  struct hostent hpstr;
#ifdef HAVE_GETHOSTBYNAME_R_3
		  struct hostent_data hdata;
#endif
		  char   buf[BUFSIZ];
		  int    herrno = 0;

		  for (i = 0; (i < 3) && (he == NULL); i++) {
#ifdef HAVE_GETHOSTBYNAME_R_3
		    if (gethostbyname_r(ascii, &hpstr, &hdata) == 0) {
		      he = &hpstr;
		    } else {
		      he = NULL;
		    }
#elif defined(HAVE_GETHOSTBYNAME_R_5)
		    he = gethostbyname_r(ascii, &hpstr, buf, sizeof(buf), &herrno);
#else
		    gethostbyname_r(ascii, &hpstr, buf, sizeof(buf), &he, &herrno);
#endif
		  }
		  if (he == NULL) {
/*			host_addr_add(List,connp->hostname, NULL);*/
			connp->err = DPS_NET_CANT_RESOLVE;
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
			DPS_FREE(uni); DPS_FREE(ascii);
#endif
			TRACE_OUT(Indexer);
			return -7;
		  }
		  memcpy((char *)&(connp->sin.sin_addr), he->h_addr, (size_t)he->h_length);
		}
#else
		{
		  struct hostent *he = NULL;
		
		  /* 
		     Note:
		     Caller must use mutex lock: 
		     gethostbyname() isn't reentrant 
		  */
		  
		  DPS_GETLOCK(Indexer, DPS_LOCK_RESOLV);
		  
		  for (i = 0; (i < 3) && (he == NULL); i++) {
		    he = gethostbyname(ascii);
		  }

		  if (he == NULL) {
/*			host_addr_add(List,connp->hostname, NULL);*/
			connp->err = DPS_NET_CANT_RESOLVE;
			DPS_RELEASELOCK(Indexer, DPS_LOCK_RESOLV);
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
			DPS_FREE(uni); DPS_FREE(ascii);
#endif
			TRACE_OUT(Indexer);
			return -8;
		  }
		  for (i = 0; he->h_addr_list[i] != NULL; i++) {
		    memcpy((char *)&(connp->sin.sin_addr), he->h_addr_list[i], (size_t)he->h_length);
		    if (DpsCheckAddr(&(connp->sin), connp->timeout) == DPS_OK) break;
		  }

		  DPS_RELEASELOCK(Indexer, DPS_LOCK_RESOLV);
		}
#endif

		host_addr_add(List, connp->hostname, &connp->sin.sin_addr);
#if defined(WITH_IDN) && !defined(APACHE1) && !defined(APACHE2)
		DPS_FREE(uni); 
		DPS_FREE(ascii);
#endif
	}else{
		if(!(Host=host_addr_find(List,connp->hostname))){
			host_addr_add(List,connp->hostname, &connp->sin.sin_addr);
		}
	}
	connp->Host = host_addr_find(List, connp->hostname);
	/*	Set port	*/
	connp->sin.sin_port = htons(connp->port);

#if defined(WITH_TRACE) && defined(DEBUG)
	fprintf(Indexer->TR, "Resolver: %s - > %s [Host.addr: %s]\n", connp->hostname, inet_ntoa(connp->sin.sin_addr),
		inet_ntoa(connp->Host->addr));
	fflush(Indexer->TR);
#endif

	TRACE_OUT(Indexer);
	return 0;
}

void DpsHostListFree(DPS_HOSTLIST *List){
	size_t i;
	
	for(i=0; i<List->nhost_addr; i++)
		DPS_FREE(List->host_addr[i].hostname);
	
	DPS_FREE(List->host_addr);
	List->nhost_addr=0;
	return;
}


