/* Copyright (C) 2003-2005 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>

#include "dps_common.h"
#include "dps_utils.h"
#include "dps_unicode.h"
#include "dps_unidata.h"
#include "dps_uniconv.h"
#include "dps_searchtool.h"
#include "dps_boolean.h"
#include "dps_xmalloc.h"
#include "dps_spell.h"
#include "dps_stopwords.h"
#include "dps_word.h"
#include "dps_vars.h"
#include "dps_db.h"
#include "dps_db_int.h"
#include "dps_url.h"
#include "dps_hash.h"
#include "dps_parsehtml.h"
#include "dps_store.h"
#include "dps_doc.h"
#include "dps_conf.h"
#include "dps_result.h"
#include "dps_log.h"
#include "dps_sgml.h"
#include "dps_mutex.h"
#include "dps_chinese.h"

#ifdef CHASEN
#include <chasen.h>
#endif

#ifdef MECAB
#include <mecab.h>
#endif

/*
#define DEBUG_CACHE
*/
/*
#define DEBUG_PHRASES
*/

typedef struct {
  size_t position;
  size_t order;
} DPS_PHR_DAT;


/********** QSORT functions *******************************/

static int cmpword(DPS_URL_CRD *s1,DPS_URL_CRD *s2){
        if (s1->coord > s2->coord) return -1;
	if (s1->coord < s2->coord) return 1;
	if (s1->url_id > s2->url_id) return 1;
	if (s1->url_id < s2->url_id) return -1;
	return 0;
}

static int cmpurlid(DPS_URL_CRD *s1,DPS_URL_CRD *s2){
	if (s1->url_id < s2->url_id) return -1;
	if (s1->url_id > s2->url_id) return 1;
	{
	  int pos1, pos2;
	  if ((pos1 = DPS_WRDPOS(s1->coord)) < (pos2 = DPS_WRDPOS(s2->coord))) return -1;
	  if (pos1 > pos2) return 1;
	}
	return 0;
}



static int DpsCmpPattern(DPS_URLCRDLIST *L, size_t i, size_t j, const char *pattern) {
  
  for(; *pattern != '\0'; pattern++) {
    switch(*pattern) {
    case 'R':
      if (L->Coords[i].coord > L->Coords[j].coord) return -1;
      if (L->Coords[i].coord < L->Coords[j].coord) return 1;
      break;
    case 'r':
      if (L->Coords[i].coord > L->Coords[j].coord) return 1;
      if (L->Coords[i].coord < L->Coords[j].coord) return -1;
      break;
    case 'P':
      if (L->Data[i].pop_rank > L->Data[j].pop_rank) return -1;
      if (L->Data[i].pop_rank < L->Data[j].pop_rank) return 1;
      break;
    case 'p':
      if (L->Data[i].pop_rank > L->Data[j].pop_rank) return 1;
      if (L->Data[i].pop_rank < L->Data[j].pop_rank) return -1;
      break;
    case 'D':
      if (L->Data[i].last_mod_time > L->Data[j].last_mod_time) return -1;
      if (L->Data[i].last_mod_time < L->Data[j].last_mod_time) return 1;
      break;
    case 'd':
      if (L->Data[i].last_mod_time > L->Data[j].last_mod_time) return 1;
      if (L->Data[i].last_mod_time < L->Data[j].last_mod_time) return -1;
      break;
    case 'I':
      {
	double m1 = L->Data[i].pop_rank, m2 = L->Data[j].pop_rank;
	m1 *= (double)L->Coords[i].coord; m2 *= (double)L->Coords[j].coord;
	if (m1 > m2) return -1;
	if (m1 < m2) return 1;
      }
	break;
    case 'i':
      {
	double m1 = L->Data[i].pop_rank, m2 = L->Data[j].pop_rank;
	m1 *= (double)L->Coords[i].coord; m2 *= (double)L->Coords[j].coord;
	if (m1 > m2) return 1;
	if (m1 < m2) return -1;
      }
	break;
    case 'A':
      {
	double m1 = L->Data[i].pop_rank * 1000.0, m2 = L->Data[j].pop_rank * 1000.0;
	m1 += (double)L->Coords[i].coord; m2 += (double)L->Coords[j].coord;
	if (m1 > m2) return -1;
	if (m1 < m2) return 1;
      }
	break;
    case 'a':
      {
	double m1 = L->Data[i].pop_rank * 1000.0, m2 = L->Data[j].pop_rank * 1000.0;
	m1 += (double)L->Coords[i].coord; m2 += (double)L->Coords[j].coord;
	if (m1 > m2) return 1;
	if (m1 < m2) return -1;
      }
	break;
    }
  }
  return 0;
}

static int DpsCmpSiteid(DPS_URLCRDLIST *L, size_t i, size_t j, const char *pattern) {

        if (L->Data[i].site_id > L->Data[j].site_id) return -1;
	if (L->Data[i].site_id < L->Data[j].site_id) return 1;
	return DpsCmpPattern(L, i, j, pattern);
}



static int cmpphr(DPS_PHR_DAT *p1, DPS_PHR_DAT *p2) {
  if (p1->position < p2->position) return -1;
  if (p1->position > p2->position) return 1;
  return 0;
}

/****************************************************/

void DpsSortSearchWordsByWeight(DPS_URL_CRD *wrd,size_t num){
	if (wrd != NULL && num > 1) DpsSort((void*)wrd, num, sizeof(*wrd), (qsort_cmp)cmpword);
	return;
}

void DpsSortSearchWordsByURL(DPS_URL_CRD *wrd,size_t num){
	if(wrd != NULL && num > 1) DpsSort((void*)wrd, num, sizeof(*wrd),( qsort_cmp)cmpurlid);
	return;
}


#define med3(func, a, b, c, L, pattern)  (func(L, a, b, pattern) < 0 ?	\
					  (func(L, b, c, pattern) < 0 ? (b) : (func(L, a, c, pattern) < 0 ? (c) : (a))) \
					  : (func(L, b, c, pattern) > 0 ? (b) : (func(L, a, c, pattern) < 0 ? (a) : (c))))



static size_t DpsPartitionSearchWordsBySite(DPS_URLCRDLIST *L, size_t p, size_t r, const char *pattern) {
  size_t i = p, j = r, m, pl, pm, pn, d;
  DPS_URL_CRD Crd;
  DPS_URLDATA Dat;

  pm = (p + r) / 2;
  pl = p;
  pn = r;
  if ( (d = (r - p)) > 160) {
    d /= 8;
    pl = med3(DpsCmpSiteid, pl, pl + d, pl + 2 * d, L, pattern);
    pm = med3(DpsCmpSiteid, pm - d, pm, pm + d, L, pattern);
    pn = med3(DpsCmpSiteid, pn - 2 * d, pn - d, pn, L, pattern);
  }
  m = med3(DpsCmpSiteid, pl, pm, pn, L, pattern);

  while (1) {
    while ((DpsCmpSiteid(L, j, m, pattern) >= 0) && (j > p)) j--;
    while ((DpsCmpSiteid(L, i, m, pattern) <= 0) && (i < r)) i++;
    if (i < j) {
      Crd = L->Coords[j];
      Dat = L->Data[j];
      L->Coords[j] = L->Coords[i];
      L->Data[j] = L->Data[i];
      L->Coords[i] = Crd;
      L->Data[i] = Dat;
    } else return j;
  }
}

/*
static size_t DpsPartitionSearchWordsBySite2(DPS_URLCRDLIST *L, size_t p, size_t r) {
  ssize_t i, j, c;
  DPS_URL_CRD Crd;
  DPS_URLDATA Dat;

  c = (p + r) / 2;
  i = j = c;

  while(1) {
    while(DpsCmpSiteid(L, j, c) >= 0 && j <= (ssize_t)r) j++;
    while(DpsCmpSiteid(L, i, c) < 0 && i >= (ssize_t)p) i--;
    if (i >= (ssize_t)p && j <= (ssize_t)r) {
      Crd = L->Coords[j];
      Dat = L->Data[j];
      L->Coords[j] = L->Coords[i];
      L->Data[j] = L->Data[i];
      L->Coords[i] = Crd;
      L->Data[i] = Dat;
    } else {
      if (i < p) {
	while (j <= (ssize_t)r) {
	  if (DpsCmpSiteid(L, j, c) < 0) {
	    Crd = L->Coords[j];
	    Dat = L->Data[j];
	    L->Coords[j] = L->Coords[c];
	    L->Data[j] = L->Data[c];
	    L->Coords[c] = Crd;
	    L->Data[c] = Dat;
	    c++;
	  }
	  j++;
	}
	return c;
      } else {
	while(i >= (ssize_t)p) {
	  if (DpsCmpSiteid(L, i, c) > 0) {
	    Crd = L->Coords[i];
	    Dat = L->Data[i];
	    L->Coords[i] = L->Coords[c];
	    L->Data[i] = L->Data[c];
	    L->Coords[c] = Crd;
	    L->Data[c] = Dat;
	    c--;
	  }
	  i--;
	}
	return c;
      }
    }
  }

}
*/



static void DpsQsortSearchWordsBySite(DPS_URLCRDLIST *L, size_t p, size_t r, const char *pattern) {

  size_t l = p, q;
  DPS_URL_CRD Crd;
  DPS_URLDATA Dat;

 DpsQsortBySiteLoop:
  if (l >= r) return;
  if (r - l < 9) {
    register size_t i, j;

    for (j = l + 1; j <= r; j++) {

      for (i = j; i > l && (DpsCmpSiteid(L, i - 1, i, pattern) > 0); i--) {
	Crd = L->Coords[i];
	Dat = L->Data[i];
	L->Coords[i] = L->Coords[i - 1];
	L->Data[i] = L->Data[i - 1];
	L->Coords[i - 1] = Crd;
	L->Data[i - 1] = Dat;
      }

    }
    return;
  }
  q = DpsPartitionSearchWordsBySite(L, l, r, pattern);
/*  q = DpsPartitionSearchWordsBySite2(L, l, r);*/
/*  fprintf(stderr, "l: %d  r: %d\n", l, r);
  fprintf(stderr, "q: %d\n", q);*/
  if (q == r) { 
    Crd = L->Coords[l];
    Dat = L->Data[l];
    L->Coords[l] = L->Coords[r];
    L->Data[l] = L->Data[r];
    L->Coords[r] = Crd;
    L->Data[r] = Dat;
    r--; 
    goto DpsQsortBySiteLoop; 
  }
/*  fprintf(stderr, "q: %d\n\n", q);*/
  DpsQsortSearchWordsBySite(L, l, q, pattern);
  l = q + 1;
  goto DpsQsortBySiteLoop;
  

}

/*#if 0*/

void DpsSortSearchWordsBySite(DPS_URLCRDLIST *L, size_t num, const char *pattern) {

  DpsQsortSearchWordsBySite(L, 0, num - 1, pattern);

}

/*#endif*/


static size_t DpsPartitionSearchWordsByPattern(DPS_RESULT *Res, DPS_URLCRDLIST *L, size_t p, size_t r, const char *pattern) {
  size_t i = p, j = r, m, pl, pm, pn, d;
  size_t Cnt = 1;
  DPS_URL_CRD Crd;
  DPS_URLDATA Dat;

  pm = (p + r) / 2;
  pl = p;
  pn = r;
  if ( (d = (r - p)) > 160) {
    d /= 8;
    pl = med3(DpsCmpPattern, pl, pl + d, pl + 2 * d, L, pattern);
    pm = med3(DpsCmpPattern, pm - d, pm, pm + d, L, pattern);
    pn = med3(DpsCmpPattern, pn - 2 * d, pn - d, pn, L, pattern);
  }
  m = med3(DpsCmpPattern, pl, pm, pn, L, pattern);

  while (1) {
    while ((DpsCmpPattern(L, j, m, pattern) >= 0) && (j > p)) j--;
    while ((DpsCmpPattern(L, i, m, pattern) <= 0) && (i < r)) i++;
    if (i < j) {
      Crd = L->Coords[j];
      Dat = L->Data[j];
      L->Coords[j] = L->Coords[i];
      L->Data[j] = L->Data[i];
      L->Coords[i] = Crd;
      L->Data[i] = Dat;
      if (Res->PerSite) {
	Cnt = Res->PerSite[j];
	Res->PerSite[j] = Res->PerSite[i];
	Res->PerSite[i] = Cnt;
      }
    } else return j;
  }
}


static void DpsQsortSearchWordsByPattern(DPS_RESULT *Res, DPS_URLCRDLIST *L, size_t p, size_t q, const char *pattern) {

  size_t l = p, r = q, c;
  size_t Cnt = 1;
  DPS_URL_CRD Crd;
  DPS_URLDATA Dat;

 DpsQsortByPatternLoop:
  if (l >= r) return;
  if (r - l < 9) {
    register size_t i, j;

    for (j = l + 1; j <= r; j++) {

      for (i = j; i > l && (DpsCmpPattern(L, i - 1, i, pattern) > 0); i--) {
	Crd = L->Coords[i];
	Dat = L->Data[i];
	L->Coords[i] = L->Coords[i - 1];
	L->Data[i] = L->Data[i - 1];
	L->Coords[i - 1] = Crd;
	L->Data[i - 1] = Dat;
	if (Res->PerSite) {
	  Cnt = Res->PerSite[i];
	  Res->PerSite[i] = Res->PerSite[i - 1];
	  Res->PerSite[i - 1] = Cnt;
	}
      }

    }
    return;
  }
  c = DpsPartitionSearchWordsByPattern(Res, L, l, r, pattern);
/*  fprintf(stderr, "l: %d  r: %d\n", l, r);
  fprintf(stderr, "q: %d\n", q);*/
  if (c == r) { 
    Crd = L->Coords[l];
    Dat = L->Data[l];
    L->Coords[l] = L->Coords[r];
    L->Data[l] = L->Data[r];
    L->Coords[r] = Crd;
    L->Data[r] = Dat;
    if (Res->PerSite) {
      Cnt = Res->PerSite[l];
      Res->PerSite[l] = Res->PerSite[r];
      Res->PerSite[r] = Cnt;
    }
    r--; 
    goto DpsQsortByPatternLoop; 
  }
/*  fprintf(stderr, "q: %d\n\n", q);*/
/*  if (c > l + 1) DpsQsortSearchWordsByPattern(Res, L, l, c - 1, pattern);*/
  DpsQsortSearchWordsByPattern(Res, L, l, c, pattern);
  l = c + 1;
  goto DpsQsortByPatternLoop;
  

}


void DpsSortSearchWordsByPattern(DPS_RESULT *Res, DPS_URLCRDLIST *L, size_t num, const char *pattern) {

  if (num > 1) DpsQsortSearchWordsByPattern(Res, L, 0, num - 1, pattern);

}


#if 0


static size_t DpsH[] = {1, 5, 19, 41, 109, 209, 505, 929, 2161, 3905, 8929, 16001, 36289, 64769};


void DpsSortSearchWordsBySite(DPS_URLCRDLIST *L, size_t num){
        register ssize_t h, i, j;
	int s = 13;
	DPS_URL_CRD Crd;
	DPS_URLDATA Dat;

	while((s > 0) && ((num / 3) < DpsH[s])) s--;
	while(s >= 0) {
	  h = DpsH[s];
	  for (j = h; j < (ssize_t)num; j++) {
	    Crd = L->Coords[j];
	    Dat = L->Data[j];

	    i = j - h;
DD4:
	    if (cmpsiteid(L, &Crd, &Dat, i) <= 0) goto DD6;
	    L->Coords[i + h] = L->Coords[i];
	    L->Data[i + h] = L->Data[i];
	    i -= h;
	    if (i >= 0) goto DD4;

DD6:
	    L->Coords[i + h] = Crd;
	    L->Data[i + h] = Dat;
	  }
	  s--;
	}

	return;
}

#endif

#if 0


void DpsSortSearchWordsByPattern(DPS_RESULT *Res, DPS_URLCRDLIST *L, size_t num, const char *pattern) {
        register ssize_t h, i, j;
	int s = 13;
	DPS_URL_CRD Crd;
	DPS_URLDATA Dat;
	size_t Cnt = 1;

	while( (s > 0) && ((num / 3) < DpsH[s])) s--;
	while(s >= 0) {
	  h = DpsH[s];
	  for (j = h; j < (ssize_t)num; j++) {
	    Crd = L->Coords[j];
	    Dat = L->Data[j];
	    if (Res->PerSite) Cnt = Res->PerSite[j];

	    i = j - h;
D4:
	    if (cmppattern(L, &Crd, &Dat, i, pattern) <= 0) goto D6;
	    L->Coords[i + h] = L->Coords[i];
	    L->Data[i + h] = L->Data[i];
	    if (Res->PerSite) Res->PerSite[i + h] = Res->PerSite[i];
	    i -= h;
	    if (i >= 0) goto D4;

D6:
	    L->Coords[i + h] = Crd;
	    L->Data[i + h] = Dat;
	    if (Res->PerSite) Res->PerSite[i + h] = Cnt;
	  }
	  s--;
	}

	return;
}

#endif




int DpsPrepare(DPS_AGENT *query, DPS_RESULT *Res) {
	DPS_CHARSET * browser_cs, *local_cs, *sys_int;
	int  ctype, ORDER = -1;
	dpsunicode_t *ustr, *nfc, *lt, *lex;
	int word_match   = DpsVarListFindInt(&query->Vars, "wm", DPS_MATCH_FULL);
	int sp   = DpsVarListFindInt(&query->Vars, "sp", 1);
	const char * txt = DpsVarListFindStr(&query->Vars, "q", "");
	const char * qlang = DpsVarListFindStr(&query->Vars, "g", NULL);
	char *ltxt;
	size_t i, wlen, llen, nphrasecmd = 0;
	char *wrd, *clex;
	dpsunicode_t *uwrd, *af_uwrd;
	DPS_CONV uni_lc, bc_uni;
	const char *lang;

	if (Res->prepared) return 0;
	
	DpsWideWordListFree(&Res->WWList);
				  
	if ((wrd = (char*)DpsMalloc(query->Conf->WordParam.max_word_len * 12 + 1)) == NULL) return 0;
	if ((uwrd = (dpsunicode_t*)DpsMalloc(sizeof(dpsunicode_t) * (query->Conf->WordParam.max_word_len + 1))) == NULL) { 
	  DPS_FREE(wrd); return 0; 
	}


	if (!(browser_cs = query->Conf->bcs)) {
		browser_cs=DpsGetCharSet("iso-8859-1");
	}
	if(!(local_cs = query->Conf->lcs)) {
		local_cs=DpsGetCharSet("iso-8859-1");
	}
	
	if (!(sys_int=DpsGetCharSet("sys-int"))) {
	        DPS_FREE(uwrd); DPS_FREE(wrd);
		return 0;
	}

	DpsConvInit(&bc_uni,browser_cs,sys_int,DPS_RECODE_HTML);
	DpsConvInit(&uni_lc,sys_int,local_cs,DPS_RECODE_HTML);
	
	llen = 14 * strlen(txt) + 1;
	ustr = (dpsunicode_t*)(DpsMalloc(sizeof(dpsunicode_t) * llen));
	if (ustr == NULL) {
	        DPS_FREE(uwrd); DPS_FREE(wrd);
		return 0;
	}
	DpsConv(&bc_uni, (char*)ustr, sizeof(dpsunicode_t) * llen, txt, llen);

	/* Create copy of query, converted into LocalCharset (for DpsTrack) */
	ltxt=(char*)DpsMalloc(llen);
	if (ltxt == NULL) {
	  DPS_FREE(uwrd); DPS_FREE(wrd); DPS_FREE(ustr);
		return 0;
	}
	DpsConv(&uni_lc, ltxt, llen, (char*)ustr, sizeof(dpsunicode_t) * llen /*bc_uni.obytes*/);
	ltxt[uni_lc.obytes]='\0';
	DpsVarListReplaceStr(&query->Vars, "q", ltxt);  /* "q-lc" was here */
	DpsLog(query, DPS_LOG_DEBUG, "Prepare query: %s, ltxt:%s", txt, ltxt);
	
	DPS_FREE(ltxt);
	
	/* Parse query and build boolean search stack*/
	DpsUniStrToLower(ustr);
	nfc = DpsUniNormalizeNFC(NULL, ustr);
	DPS_FREE(ustr);
	switch(browser_cs->family) {
	case DPS_CHARSET_CHINESE_SIMPLIFIED:
	case DPS_CHARSET_CHINESE_TRADITIONAL: lang = "zh"; break;
	case DPS_CHARSET_JAPANESE: lang = "ja"; break;
	case DPS_CHARSET_THAI: lang = "th"; break;
	case DPS_CHARSET_UNICODE: lang = qlang; break;
	default: lang = "";
	}

	DpsLog(query, DPS_LOG_EXTRA, "Segment lang: %s", lang);

	ustr = DpsUniSegment(query, nfc, lang);

	lex = (ustr != NULL) ? DpsUniGetSepToken(ustr, &lt , &ctype) : NULL;
	while(lex){
	  wlen = lt - lex;
	  memcpy(uwrd, lex, (dps_min(wlen, query->Conf->WordParam.max_word_len)) * sizeof(dpsunicode_t));
	  uwrd[dps_min(wlen, query->Conf->WordParam.max_word_len)] = 0;
	  DpsConv(&uni_lc, wrd, query->Conf->WordParam.max_word_len * 12,(char*)uwrd, sizeof(uwrd[0])*(wlen+1));
	  clex = DpsTrim(wrd, " \t\r\n");

	  DpsLog(query, DPS_LOG_DEBUG, "wrd {%d}: %s", wlen, wrd);
			
	  if (Res->nitems >= Res->mitems) {
	    Res->mitems += DPS_MAXSTACK;
	    Res->items = (DPS_STACK_ITEM*)DpsRealloc(Res->items, Res->mitems * sizeof(DPS_STACK_ITEM));
	    if (Res->items == NULL) {
	      DPS_FREE(uwrd); DPS_FREE(wrd); DPS_FREE(ustr); DPS_FREE(Res->items);
	      return 0;
	    }
	  }
	  if (DPS_UNI_CTYPECLASS(ctype) != DPS_UNI_BUKVA) {
/*	    if (nphrasecmd & 1)*/
	    size_t ULen = DpsUniLen(uwrd);
	      for (i = 0; i < ULen; i++) {
		        switch(uwrd[i]) {
			case '&':
			case '+':
			  Res->items[Res->nitems].cmd = DPS_STACK_AND;
			  break;
			case '|':
			  Res->items[Res->nitems].cmd = DPS_STACK_OR;
			  break;
			case '~':
			  Res->items[Res->nitems].cmd = DPS_STACK_NOT;
			  break;
			case '(':
			  Res->items[Res->nitems].cmd = DPS_STACK_LEFT;
			  break;
			case ')':
			  Res->items[Res->nitems].cmd = DPS_STACK_RIGHT;
			  break;
			case '"':
			case '\'':
			  Res->items[Res->nitems].cmd = DPS_STACK_PHRASE;
			  Res->phrase = 1;
			  nphrasecmd++;
			  break;
			default: continue;
			}
			Res->items[Res->nitems].arg=0;
			Res->nitems++;
			Res->ncmds++;
	      }
	  } else if (!(nphrasecmd & 1) && (strcasecmp(clex, "AND") == 0)) {
	    Res->items[Res->nitems].cmd = DPS_STACK_AND;
	    Res->items[Res->nitems].arg=0;
	    Res->nitems++;
	    Res->ncmds++;
	  } else if (!(nphrasecmd & 1) && (strcasecmp(clex, "OR") == 0)) {
	    Res->items[Res->nitems].cmd = DPS_STACK_OR;
	    Res->items[Res->nitems].arg=0;
	    Res->nitems++;
	    Res->ncmds++;
	  } else if (!(nphrasecmd & 1) && (strcasecmp(clex, "NOT") == 0)) {
	    Res->items[Res->nitems].cmd = DPS_STACK_NOT;
	    Res->items[Res->nitems].arg=0;
	    Res->nitems++;
	    Res->ncmds++;
	  } else if (!(nphrasecmd & 1) && (strcasecmp(clex, "~") == 0)) {
	    Res->items[Res->nitems].cmd = DPS_STACK_NOT;
	    Res->items[Res->nitems].arg=0;
	    Res->nitems++;
	    Res->ncmds++;
	  } else {
			int addwrd=1;

			ORDER++;
			Res->items[Res->nitems].cmd=DPS_STACK_WORD;
			Res->items[Res->nitems].arg = ORDER /*Res->WWList.nuniq*/;  /* 1L << (Res->WWList.nuniq); */
			Res->nitems++;
			if (Res->nitems >= Res->mitems) {
			  Res->mitems += DPS_MAXSTACK;
			  Res->items = (DPS_STACK_ITEM*)DpsRealloc(Res->items, Res->mitems * sizeof(DPS_STACK_ITEM));
			  if (Res->items == NULL) {
			    DPS_FREE(uwrd); DPS_FREE(wrd); DPS_FREE(ustr); DPS_FREE(Res->items);
			    return 0;
			  }
			}

			if(word_match==DPS_MATCH_FULL){
				/* Check stopword only when full word         */
				/* Substring searches should not exclude them */
				if(DpsStopListFind(&query->Conf->StopWords, uwrd, qlang) ||
						(query->Conf->WordParam.min_word_len>wlen)||
						(query->Conf->WordParam.max_word_len<wlen)){
				  DPS_WIDEWORD OWord;
				
				  OWord.len=strlen(wrd);
				  OWord.order = ORDER /*Res->WWList.nuniq*/;
				  OWord.count=0;
				  OWord.crcword=DpsStrHash32(wrd);
				  OWord.word=wrd;
				  OWord.uword=uwrd;
				  OWord.origin = DPS_WORD_ORIGIN_STOP;
				  DpsWideWordListAdd(&Res->WWList, &OWord);
				  addwrd=0;
				}
			}
			if(Res->WWList.nuniq >= DPS_MAXWORDPERQUERY-1){
				addwrd=0;
			}
			if(addwrd){
				DPS_WIDEWORD OWord;
				DPS_WIDEWORDLIST * forms;
				
				OWord.len=strlen(wrd);
				OWord.order = ORDER /*Res->WWList.nuniq*/;
				OWord.count=0;
				OWord.crcword=DpsStrHash32(wrd);
				OWord.word=wrd;
				OWord.uword=uwrd;
				OWord.origin = DPS_WORD_ORIGIN_QUERY;
				DpsWideWordListAdd(&Res->WWList, &OWord);
				
				if(sp && (forms = DpsAllForms(query, &OWord)) ) {
					DPS_WIDEWORD FWord;
					size_t frm;
					for(frm=0;frm<forms->nwords;frm++){
						DpsConv(&uni_lc,wrd,12*query->Conf->WordParam.max_word_len,
							(char*)(forms->Word[frm].uword),
							sizeof(forms->Word[frm].uword[0])*(DpsUniLen(forms->Word[frm].uword)+1));
						FWord.len=strlen(wrd);
						FWord.order = ORDER; /*Res->WWList.nuniq;*/
						FWord.count=0;
						FWord.crcword=DpsStrHash32(wrd);
						FWord.word=wrd;
						FWord.uword=forms->Word[frm].uword;
						FWord.origin = forms->Word[frm].origin;
						
						DpsWideWordListAdd(&Res->WWList,&FWord);
/*						DpsLog(query, DPS_LOG_DEBUG, "Word form: [%d] %s", FWord.origin, wrd);*/
					}
					DpsWideWordListFree(forms);
					DPS_FREE(forms);
				}

				if (query->Flags.use_accentext) {
				  af_uwrd = DpsUniAccentStrip(uwrd);
				  if (DpsUniStrCmp(af_uwrd, uwrd) != 0) {
				    DpsConv(&uni_lc, wrd, query->Conf->WordParam.max_word_len * 12, (char*)af_uwrd, 
					    sizeof(af_uwrd[0])*(DpsUniLen(af_uwrd) + 1));
				    clex = DpsTrim(wrd, " \t\r\n");

				    OWord.len = strlen(wrd);
				    OWord.order = ORDER /*Res->WWList.nuniq*/;
				    OWord.count = 0;
				    OWord.crcword = DpsStrHash32(wrd);
				    OWord.word = wrd;
				    OWord.uword = af_uwrd;
				    OWord.origin = DPS_WORD_ORIGIN_ACCENT;
				    DpsWideWordListAdd(&Res->WWList, &OWord);
				
				    if(sp && (forms = DpsAllForms(query, &OWord))) {
					DPS_WIDEWORD FWord;
					size_t frm;
					for(frm=0;frm<forms->nwords;frm++){
						DpsConv(&uni_lc,wrd,12*query->Conf->WordParam.max_word_len,
							(char*)(forms->Word[frm].uword),
							sizeof(forms->Word[frm].uword[0])*(DpsUniLen(forms->Word[frm].uword)+1));
						FWord.len=strlen(wrd);
						FWord.order = ORDER; /*Res->WWList.nuniq;*/
						FWord.count=0;
						FWord.crcword=DpsStrHash32(wrd);
						FWord.word=wrd;
						FWord.uword=forms->Word[frm].uword;
						FWord.origin = forms->Word[frm].origin;
						
						DpsWideWordListAdd(&Res->WWList,&FWord);
/*						DpsLog(query, DPS_LOG_DEBUG, "Word form: [%d] %s", FWord.origin, wrd);*/
					}
					DpsWideWordListFree(forms);
					DPS_FREE(forms);
				    }
				  }
				  DPS_FREE(af_uwrd);
				}


			}
		}
		lex = DpsUniGetSepToken(NULL, &lt, &ctype);
	}
	if (nphrasecmd & 1) {
	  Res->items[Res->nitems].cmd = DPS_STACK_PHRASE;
	  Res->items[Res->nitems].arg=0;
	  Res->nitems++;
	  Res->ncmds++;
	  if (Res->nitems >= Res->mitems) {
	    Res->mitems += DPS_MAXSTACK;
	    Res->items = (DPS_STACK_ITEM*)DpsRealloc(Res->items, Res->mitems * sizeof(DPS_STACK_ITEM));
	    if (Res->items == NULL) {
	      DPS_FREE(uwrd); DPS_FREE(wrd); DPS_FREE(ustr); DPS_FREE(Res->items);
	      return 0;
	    }
	  }
	}
	Res->orig_nitems = Res->nitems;
	DPS_FREE(ustr); DPS_FREE(uwrd); DPS_FREE(wrd);
	Res->prepared = 1;
	return(0);
}


static unsigned DpsBitCntTable[256] = {
 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
 4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};


static int DpsBitsCount(unsigned int x) {
  return DpsBitCntTable[ x & 0xFF] +DpsBitCntTable[(x >> 8) & 0xFF] + DpsBitCntTable[(x >> 16) & 0xFF] + DpsBitCntTable[(x >> 24)]; 
}



static uint4 DpsCalcCosineWeight(unsigned int *R, double x, unsigned int *D, size_t nw, size_t ns, size_t nadd) {
  register double /*x = 0.0,*/ y = 0.0, xy = 0.0;
  register size_t i;
#ifdef FULL_RELEVANCY
  register size_t j;
#endif

  if (x == 0.0) return (uint4)0;

  for (i = 0; i < nadd; i++) {
    if (D[i] > R[i]) {
      y += (double) (D[i] - R[i]);
    } else {
      y += (double) (R[i] - D[i]);
    }
  }

/*  fprintf(stderr, "y: %lf  ", y);*/

#ifdef FULL_RELEVANCY

  for (i = 0; i < nw; i++) 
    for (j = 0; j < ns; j++) {
      xy += (double) DpsBitsCount(R[nadd + i * 256 + j] ^ D[nadd + i * 256 + j]);
    }
#else
  for (i = 0; i < nw; i++) {
    xy += (double)DpsBitsCount(R[nadd + i] ^ D[nadd + i]);
/*    fprintf(stderr, "[%d %x:%x]", i, R[nadd + i], D[nadd + i]);*/
  }
#endif

/*  fprintf(stderr, "\n1. x: %lf  xy: %lf  y: %lf\n", x, xy, y);*/

  y = 1.0 / log(2.8 + y);
/*  fprintf(stderr, "2. x: %lf  xy: %lf  y: %lf\n", x, xy, y);*/

/*  y = (xy + y) / (x + y);*/
/*  y = 1.0 - (xy + y) / (x + y);*/
  y *= (xy / x);
/*  fprintf(stderr, "3. x: %lf  xy: %lf  y: %lf\n", x, xy, y);*/

  if (y < 0.0) return (uint4)0;
  return (uint4)floor(100000.0 * y);

/*  fprintf(stderr, "rating: %u\n\n", (uint4)floor(100000.0 * (1.0 - (xy + y) / (x + y) ) ));*/

/*  return (uint4)floor(100000.0 * (1.0 - (xy + y) / (x + y)));*/
/*  return (uint4)floor( 100000.0 * ( xy + y) / (x + y) );*/
  
}


static int DpsOriginWeight(int origin) {  /* Weight for origin can be from 1 to 15 */
  switch(origin) {
  case DPS_WORD_ORIGIN_QUERY: return 8;
  case DPS_WORD_ORIGIN_SPELL: return 2;
  case DPS_WORD_ORIGIN_ACCENT: return 4;
  case DPS_WORD_ORIGIN_SYNONYM: return 1;
  }
  return 0;
}

static unsigned int CalcAverageWordDistance(DPS_PHR_DAT *phr, size_t num, size_t nuniq, int need_sort) {
  register size_t i, n = 0;
  register uint8 sum = 0;

  if (num < 2) return 1000 + phr[0].position;
/*  if (need_sort) return 1;*/
  if (need_sort) DpsSort((void*)phr, num, sizeof(DPS_PHR_DAT), (qsort_cmp)cmpphr);
  if (nuniq > 1) {
    register size_t j = 0;
    for (i = 1; i < num; ) {
      while ((i < num) && (phr[i].order == phr[j].order)) i++;
      if ((i < num) && (phr[i].order ==  phr[j].order + 1)) {
	sum += (phr[i].position - phr[i-1].position);
	n++;
      }
      j = i++;
    }
  } else {
    for (i = 1; i < num; i++) {
      sum += (phr[i].position - phr[i-1].position);
      n++;
    }
  }
  return (n > nuniq) ? (sum / n) : (1000 + phr[0].position)/*(nuniq * 100)*/ ;
}


static int Check(DPS_PHR_DAT *phr, size_t *gap, size_t phr_n, size_t i, size_t maxord, size_t arg) {
  size_t curpos = phr[i].position;
  size_t Cgap = gap[phr[i].order];
  size_t needord = phr[i].order + Cgap;
  register size_t z;

#ifdef DEBUG_PHRASES
  fprintf(stderr, "i:%2d arg: %d  phr[i].order:%d  curpos: %4d  Cgap: %d  needord: %d\n", i, arg, phr[i].order, curpos, Cgap, needord);
#endif

  if (phr[i].order != arg) {
    return 0;
  }
  if (Cgap == 0) {
    return 1;
  }
  if (needord >= maxord) {
    return 1;
  }

  for (z = i + 1; (z < phr_n) && (phr[z].position - curpos <= Cgap /*2*/); z++) {
    if ((phr[z].order == needord) && (phr[z].position == curpos + Cgap/*1*/)) {
      return 1;
    }
  }
  return 0;
}

#ifndef FULL_RELEVANCY
#define WF_ADD(sec)    (wf[(sec)] << (4 + 4 * (((sec) - 1) % 6)))
#endif

#define DPS_N_ADD  2


void DpsGroupByURL(DPS_AGENT *query, DPS_RESULT *Res) {
	DPS_STACK_ITEM *temp_items;
	size_t	i, j = 0, D_size, nitems = Res->nitems, nstopwords = 0;
	size_t  *count, *gap, count_size = Res->WWList.nuniq * sizeof(size_t);
	int search_mode = DpsSearchMode(DpsVarListFindStr(&query->Vars, "m", "all"));
	int orig_search_mode = DpsSearchMode(DpsVarListFindStr(&query->Vars, "orig_m", "all"));
	DPS_URL_CRD *Crd = Res->CoordList.Coords;
	size_t nsections = DpsVarListFindInt(&query->Vars, "NumSections", 256);
	int wf[256];
	int add_cmd = DPS_STACK_AND, inphrase;
	unsigned int *R, *D;
	DPS_PHR_DAT *phr;
	size_t phr_n = 0, phr_m = 256;
	size_t pharg, x, y, ya, z, ph_start, ph_end, curpos;
	double Rbc;

	if(!Res->CoordList.ncoords) return;

/*#ifdef FULL_RELEVANCY*/
	DpsWeightFactorsInit(DpsVarListFindStr(&query->Vars, "wf", ""), wf);
/*#endif*/

	phr = (DPS_PHR_DAT*)DpsMalloc(phr_m * sizeof(DPS_PHR_DAT));
	if (phr == NULL) return;
	if ((count = (size_t*)DpsXmalloc(count_size)) == NULL) {
	  DPS_FREE(phr);
	  return;
	}
	if ((gap = (size_t*)DpsXmalloc(count_size)) == NULL) {
	  DPS_FREE(count);  DPS_FREE(phr);
	  return;
	}
	D_size = (Res->WWList.nwords 
#ifdef FULL_RELEVANCY
		  * 256 
#endif
		  + DPS_N_ADD) * sizeof(unsigned int) + 1;
	if ((R = (unsigned int*)DpsXmalloc(D_size)) == NULL) {
	  DPS_FREE(gap);  DPS_FREE(count);DPS_FREE(phr);
	  return;
	}
	if ((D = (unsigned int*)DpsXmalloc(D_size)) == NULL) {
	  DPS_FREE(gap); DPS_FREE(count); DPS_FREE(R);DPS_FREE(phr);
	  return;
	}
	if ((temp_items = (DPS_STACK_ITEM*)DpsMalloc((2 * Res->nitems + 1) * sizeof(DPS_STACK_ITEM))) == NULL) {
	  DPS_FREE(gap); DPS_FREE(count); DPS_FREE(R); DPS_FREE(D);DPS_FREE(phr);
	  return;
	}

	for(i = 0; i < Res->WWList.nwords; i++) {
	  if (Res->WWList.Word[i].origin == DPS_WORD_ORIGIN_STOP) {
	    gap[Res->WWList.Word[i].order] = count[Res->WWList.Word[i].order] = 1;
	    nstopwords++;
	  }
	}

	for (i = 0; i <  Res->WWList.nuniq; i++) {
	  if (gap[i] == 0) {
	    gap[i]++;
	    for (z = i + 1; z < Res->WWList.nuniq; z++) if (gap[z] != 0) gap[i]++; else break;
	    
	  } else {
	    gap[i] = 0;
	    
	  }
	}
	gap[Res->WWList.nuniq - 1] = 0;

#ifdef DEBUG_PHRASES
				    {
				      size_t ii;
				      for (ii = 0; ii < Res->WWList.nuniq; ii++) {
					fprintf(stderr, "Gap[%d]:%d\n", ii, gap[ii]);
				      }
				    }
#endif
        R[0] = (Res->WWList.nuniq == 1) ? 0 : 1; /* Res->WWList.nuniq - 1;*/ /* was: 0 */
	R[1] = 64;

	Rbc = (double)0.0;

#ifdef FULL_RELEVANCY
	for (j = 0; j < nsections; j++)
	  for(i = 0; i < Res->WWList.nwords; i++) {
	    R[DPS_N_ADD + 256 * i + j] = ((wf[j]) << 4) | DpsOriginWeight(Res->WWList.Word[i].origin);
	    Rbc += (double)DpsBitsCount(R[DPS_N_ADD + 256 * i + j]);
	  }

	D[DPS_N_ADD + 256 * DPS_WRDNUM(Crd[0].coord) + DPS_WRDSEC(Crd[0].coord)] = 
	  (wf[DPS_WRDSEC(Crd[0].coord)] << 4) | DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[0].coord)].origin);
#else
	for(i = 0; i < Res->WWList.nwords; i++) {
	  for (j = 0; j < nsections; j++)
	    R[DPS_N_ADD + i] |= WF_ADD(j);
	  R[DPS_N_ADD + i] |= DpsOriginWeight(Res->WWList.Word[i].origin);
	  Rbc += (double)DpsBitsCount(R[DPS_N_ADD + i]);
	}
		
	D[DPS_N_ADD + DPS_WRDNUM(Crd[0].coord)] |= WF_ADD(DPS_WRDSEC(Crd[0].coord)) | 
	  DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[0].coord)].origin);
#endif

	if (1 /*Res->ncmds > 0 || search_mode == DPS_MODE_BOOL*/) {
/*	  fprintf(stderr, "Res->ncmds: %d\n", Res->ncmds);*/
	  switch(orig_search_mode) {
	  case DPS_MODE_ANY: add_cmd = DPS_STACK_OR; break;
	  case DPS_MODE_BOOL:
	  case DPS_MODE_ALL: add_cmd = DPS_STACK_AND; break;
	  }
	  j = 1;
	  temp_items[0] = Res->items[0];
	  if (Res->items[0].cmd == DPS_STACK_PHRASE) inphrase = 0; else inphrase = 1;

	  for (i = 1; i < Res->nitems; i++) {
	    if (Res->items[i].cmd == DPS_STACK_PHRASE) {
	      if ((inphrase != 0) && (temp_items[j-1].cmd == DPS_STACK_WORD)) {
		temp_items[j].cmd = add_cmd;
		temp_items[j++].arg = 0;
	      }
	      inphrase = ((inphrase + 1) & 1);
	    }
#ifdef DEBUG_PHRASES
	    fprintf(stderr, "inprase: %d   cmd.i: %d  cmd.j: %d  arg.j:%d\n", inphrase, Res->items[i].cmd, 
		    temp_items[j-1].cmd, temp_items[j-1].arg );
#endif
	    if ((j != 0) && (inphrase != 0)) {

	      if (Res->items[i].cmd == DPS_STACK_PHRASE) gap[temp_items[j-1].arg] = 0;

	      if ( (Res->items[i].cmd == DPS_STACK_WORD) && 
		((temp_items[j-1].cmd == DPS_STACK_WORD) || (temp_items[j-1].cmd == DPS_STACK_PHRASE)) ) {

		temp_items[j].cmd = add_cmd;
		temp_items[j++].arg = 0;
	      }
	      if ( (Res->items[i].cmd == DPS_STACK_NOT) && 
		((temp_items[j-1].cmd == DPS_STACK_WORD) || (temp_items[j-1].cmd == DPS_STACK_PHRASE)) ) {

		temp_items[j].cmd = add_cmd;
		temp_items[j++].arg = 0;
	      }
	    }
	    temp_items[j++] = Res->items[i];
	  }
	  search_mode = DPS_MODE_BOOL;
	  Res->nitems = nitems = j;
	}


	D[1] = phr[phr_n].position = DPS_WRDPOS(Crd[0].coord);
	phr[phr_n].order = Res->WWList.Word[DPS_WRDNUM(Crd[0].coord)].order;
	count[phr[phr_n].order]++;
#ifdef DEBUG_PHRASES
	fprintf(stderr, "url_id: %d, phr_n:%02d  order:%d  position:%d\n", Crd[0].url_id, phr_n, phr[phr_n].order, phr[phr_n].position);
#endif
	phr_n++;


	j = 0;
	for(i = 1; i < Res->CoordList.ncoords; i++){
		/* Group by url_id */
		if(Crd[j].url_id==Crd[i].url_id){
		        /* Same document */

		  D[1] += (phr[phr_n].position = DPS_WRDPOS(Crd[i].coord));
		  phr[phr_n].order = Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].order;
		  count[phr[phr_n].order]++;
#ifdef DEBUG_PHRASES
	fprintf(stderr, "url_id: %d, phr_n:%02d  order:%d  position:%d\n", Crd[j].url_id, phr_n, phr[phr_n].order, phr[phr_n].position);
#endif
			if (++phr_n >= phr_m) {
			  phr_m += 256;
			  phr = (DPS_PHR_DAT*)DpsRealloc(phr, phr_m * sizeof(DPS_PHR_DAT));
			  if (phr == NULL) goto DpsRex;
			}
#ifdef FULL_RELEVANCY
			D[DPS_N_ADD + 256 * DPS_WRDNUM(Crd[i].coord) + DPS_WRDSEC(Crd[i].coord)] |= 
			  (wf[DPS_WRDSEC(Crd[i].coord)] << 4) | DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].origin);
#else
			D[DPS_N_ADD + DPS_WRDNUM(Crd[i].coord)] |= WF_ADD(DPS_WRDSEC(Crd[i].coord)) | 
			  DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].origin);
#endif
		}else{
			/* Next document */

		  if (Res->phrase > 0 && phr_n > 1) DpsSort((void*)phr, phr_n, sizeof(DPS_PHR_DAT), (qsort_cmp)cmpphr);
		        switch(search_mode) {
			case DPS_MODE_BOOL:
/* [[[[[[[[[[[[[[ */
			        for(z = 0; z < nitems - 1; z++) {
				  if (temp_items[z].cmd == DPS_STACK_PHRASE) { /* we assume correct stack at this moment,
										  i.e. no empty phrases nor illegal operations order */
				    pharg = 1;
				    ph_start = z + 1;
				    while(gap[temp_items[ph_start].arg] == 0 && ph_start < nitems-2) { ph_start++; nstopwords--; }
				    for (ph_end = ph_start; ph_end < nitems && temp_items[ph_end].cmd == DPS_STACK_WORD; ph_end++) {
				      if (count[temp_items[ph_end].arg] == 0) {
					pharg = 0;
				      }
				    }
				    if (pharg && (phr_n >= (ph_end - ph_start) || (phr_n + nstopwords) == (ph_end - ph_start) )) {
				      pharg = 0;
				      for (y = 0; y < phr_n; y++) {
					pharg = 0;
					for (x = ph_start, ya = 0; x < ph_end && (y + ya) < phr_n;) {
					  curpos = phr[y + ya].position;
					  pharg = Check(phr, gap, phr_n, y + ya, Res->WWList.nuniq, temp_items[x].arg );
					  if (pharg == 0) break;
/*					  {
					    size_t ii;
					    for (ii = 0; ii < Res->WWList.nuniq; ii++) {
					      fprintf(stderr, "gap[%d]:%d\n", ii, gap[ii]);
					    }
					  }
*/
					  if (gap[temp_items[x].arg] == 0) {
					    x++;
					  } else {
					    x += gap[temp_items[x].arg];
					  }
					  for(ya++; (y + ya < phr_n) && (curpos == phr[y + ya].position); ya++);
					}
					if (pharg) break;
				      }
				    } else pharg = 0;
				    for (y = z + 1/*ph_start*/; y < ph_end; y++) count[temp_items[y].arg] = pharg;
				    for (z++; z < nitems && (temp_items[z].cmd != DPS_STACK_PHRASE); z++);
				  }
				}
/* ]]]]]]]]]]]]]] */
				if(DpsCalcBoolItems(temp_items, nitems, count) > 0) {
				  D[0] = CalcAverageWordDistance(phr, phr_n,  Res->WWList.nuniq, 
								 (Res->phrase == 0 && Res->WWList.nwords > 1) ? 1 : 0 );
				  D[1] /= phr_n;
				  Crd[j].coord = DpsCalcCosineWeight(R, Rbc, D, Res->WWList.nwords, 
#ifdef FULL_RELEVANCY
									   nsections,
#else
									   0,
#endif
									   DPS_N_ADD);
					j++;
				}
				break;
			case DPS_MODE_ALL:
			        for (z = 0; z < Res->WWList.nuniq; z++) if (count[z] == 0) break;
				if (z < Res->WWList.nuniq && count[z] == 0) break;
			default:
 			                D[0] = CalcAverageWordDistance(phr, phr_n, Res->WWList.nuniq,
								       (Res->phrase == 0 && Res->WWList.nwords > 1) ? 1 : 0 );
					D[1] /= phr_n;
				        Crd[j].coord = DpsCalcCosineWeight(R, Rbc, D, Res->WWList.nwords, 
#ifdef FULL_RELEVANCY
									   nsections, 
#else
									   0,
#endif
									   DPS_N_ADD);
					j++;
			}


			Crd[j]=Crd[i];

			bzero((void*)D, D_size);
			bzero((void*)count, count_size);
			{ 
			  register size_t ii;
			  for(ii = 0; ii < Res->WWList.nwords; ii++) {
			    if (Res->WWList.Word[ii].origin == DPS_WORD_ORIGIN_STOP) count[Res->WWList.Word[ii].order] = 1;
			  }
			}
			phr_n = 0;
			D[1] = phr[phr_n].position = DPS_WRDPOS(Crd[i].coord);
			phr[phr_n].order = Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].order;
			count[phr[phr_n].order]++;
#ifdef DEBUG_PHRASES
	fprintf(stderr, "url_id: %d, phr_n:%02d  order:%d  position:%d [next]\n", Crd[i].url_id, phr_n, phr[phr_n].order, phr[phr_n].position);
#endif
			phr_n++;
#ifdef FULL_RELEVANCY
			D[DPS_N_ADD + 256 * DPS_WRDNUM(Crd[i].coord) + DPS_WRDSEC(Crd[i].coord)] = 
			  (wf[DPS_WRDSEC(Crd[i].coord)] << 4) | DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].origin);
#else
			D[DPS_N_ADD + DPS_WRDNUM(Crd[i].coord)] |= WF_ADD(DPS_WRDSEC(Crd[i].coord)) | 
			  DpsOriginWeight(Res->WWList.Word[DPS_WRDNUM(Crd[i].coord)].origin);
#endif
		}
	}

	/* Check last word */

/*	if (Res->phrase > 0 && phr_n > 1) DpsSort((void*)phr, phr_n, sizeof(DPS_PHR_DAT), (qsort_cmp)cmpphr);*/
	D[0] = CalcAverageWordDistance(phr, phr_n, Res->WWList.nuniq, 1);
	D[1] /= phr_n;

	switch(search_mode) {
	case DPS_MODE_BOOL:
/* [[[[[[[[[[[[[[ */
			        for(z = 0; z < nitems - 1; z++) {
				  if (temp_items[z].cmd == DPS_STACK_PHRASE) { /* we assume correct stack at this moment,
										  i.e. no empty phrases nor illegal operations order */
				    pharg = 1;
				    ph_start = z + 1;
				    while(gap[temp_items[ph_start].arg] == 0 && ph_start < nitems-2) { ph_start++; nstopwords--; }
				    for (ph_end = ph_start; ph_end < nitems && temp_items[ph_end].cmd == DPS_STACK_WORD; ph_end++) {
				      if (count[temp_items[ph_end].arg] == 0) {
					pharg = 0;
				      }
				    }
#ifdef DEBUG_PHRASES
				    {
				      size_t ii;
				      for (ii = 0; ii < Res->WWList.nuniq; ii++) {
					fprintf(stderr, "gap[%d]:%d\n", ii, gap[ii]);
				      }
				    }
				    fprintf(stderr, "ph_start: %d  ph_end: %d  phr_n: %d\n", ph_start, ph_end, phr_n);
				    fprintf(stderr, "pharg1: %d\n", pharg);
#endif
				    if (pharg && (phr_n >= (ph_end - ph_start) || (phr_n + nstopwords) == (ph_end - ph_start) )) {
				      pharg = 0;
				      for (y = 0; y < phr_n; y++) {
					pharg = 0;
#ifdef DEBUG_PHRASES
					fprintf(stderr, "ph_start: %d  ph_end: %d  phr_n: %d\n", ph_start, ph_end, phr_n);
#endif
					for (x = ph_start, ya = 0; x < ph_end && (y + ya) < phr_n;) {
					  curpos = phr[y + ya].position;
					  pharg = Check(phr, gap, phr_n, y + ya, Res->WWList.nuniq, temp_items[x].arg );
#ifdef DEBUG_PHRASES
					  fprintf(stderr, "pharg: %d\n", pharg);
#endif
					  if (pharg == 0) break;
					  if (gap[temp_items[x].arg] == 0) {
					    x++;
					  } else {
					    x += gap[temp_items[x].arg];
					  }
					  for(ya++; (y + ya < phr_n) && (curpos == phr[y + ya].position); ya++);
					}
					if (pharg) break;
				      }
				    } else pharg = 0;
#ifdef DEBUG_PHRASES
				    fprintf(stderr, "pharg2: %d\n", pharg);
#endif
				    for (y = z + 1/*ph_start*/; y < ph_end; y++) count[temp_items[y].arg] = pharg;
				    for (z++; z < nitems && (temp_items[z].cmd != DPS_STACK_PHRASE); z++);
				  }
				}
/* ]]]]]]]]]]]]]] */
#ifdef DEBUG_PHRASES
	  fprintf(stderr, "nuniq: %d\n", Res->WWList.nuniq);
	  for (z = 0; z < Res->WWList.nuniq; z++) fprintf(stderr, "count[%d]: %d\n", z, count[z]);
#endif

	  if (DpsCalcBoolItems(temp_items, nitems, count) > 0) {
	    Res->CoordList.ncoords = j + 1;
	    Crd[j].coord = DpsCalcCosineWeight(R, Rbc, D, Res->WWList.nwords, 
#ifdef FULL_RELEVANCY
					       nsections, 
#else
					       0,
#endif
					       DPS_N_ADD);
	  } else {
	    Res->CoordList.ncoords = j;
	  }

/*	  fprintf(stderr, "j: %d ncoords: %d\n", j, Res->CoordList.ncoords);*/
	  break;
	case DPS_MODE_ALL:
/*	  fprintf(stderr, "nuniq: %d\n", Res->WWList.nuniq);
	  for (z = 0; z < Res->WWList.nuniq; z++) fprintf(stderr, "count[%d]: %d\n", z, count[z]);
*/
	  for (z = 0; z < Res->WWList.nuniq; z++) if (count[z] == 0) break;
	  Res->CoordList.ncoords = (z < Res->WWList.nuniq && count[z] == 0) ? j : j + 1;
	  break;
	default:
	  Crd[j].coord = DpsCalcCosineWeight(R, Rbc, D, Res->WWList.nwords, 
#ifdef FULL_RELEVANCY
					     nsections, 
#else
					     0,
#endif
					     1);
	  Res->CoordList.ncoords = ( Crd[j].coord ) ? j + 1 : j;
	}
DpsRex:
	DPS_FREE(temp_items);
	DPS_FREE(D);
	DPS_FREE(R);
	DPS_FREE(count);
	DPS_FREE(gap);
	DPS_FREE(phr);
	return;
}


void DpsGroupBySite(DPS_AGENT *query, DPS_RESULT *Res){	
	register size_t	i, j = 0, cnt = 1; 
	urlid_t Doc_site;
	DPS_URL_CRD *Crd = Res->CoordList.Coords;
	DPS_URLDATA *Dat = Res->CoordList.Data;
	
	if(!Res->CoordList.ncoords) return;
	if ((Res->PerSite = (size_t*)DpsMalloc(sizeof(*Res->PerSite) * Res->CoordList.ncoords)) == NULL) return;

	Doc_site = Dat[0].site_id;

	for(i = 1; i < Res->CoordList.ncoords; i++){
		/* Group by site_id */
		if(Doc_site == Dat[i].site_id){
		/* Same site */
		  /* adjust document rating rating according nuber of documents from site, if we need this */
		  cnt++;
		}else{
			/* Next site */
			Res->PerSite[j] = cnt;
			cnt = 1;
		        j++;
			Doc_site = Dat[i].site_id;
			Crd[j] = Crd[i];
			Dat[j] = Dat[i];
		}
	}
	Res->PerSite[j] = cnt;
	Res->CoordList.ncoords = j + 1;

	return;
}

typedef struct{
	/*char	phr_len;*/
	uint4	count;
	uint4	weight;
} DPS_PHR_PAR;



int DpsParseQueryString(DPS_AGENT * Agent,DPS_VARLIST * vars,char * query_string){
	char * tok, *lt;
	size_t len;
	char *str = (char *)DpsMalloc((len = strlen(query_string)) + 7);
	char *qs = (char*)DpsStrdup(query_string);
	char qname[256];

	if ((str == NULL) || qs == NULL) {
	  DPS_FREE(str);
	  DPS_FREE(qs);
	  return 1;
	}

	Agent->nlimits = 0;

	DpsSGMLUnescape(qs);
	
	tok = dps_strtok_r(qs, "&", &lt);
	while(tok){
		char empty[]="";
		char * val;
		const char * lim;
		
		if((val=strchr(tok,'='))){
			*val='\0';
			val++;
		}else{
			val=empty;
		}
		DpsUnescapeCGIQuery(str,val);
		DpsVarListReplaceStr(vars,tok,str);
		dps_snprintf(qname, 256, "query.%s", tok);
		DpsVarListReplaceStr(vars, qname, str);

		sprintf(str,"Limit-%s",tok);
		if((lim=DpsVarListFindStr(vars,str,NULL))){
			int ltype=0;
			const char * type, * fname = NULL;
			char * llt;
			strncpy(str, lim, len);
			
			if((type = dps_strtok_r(str, ":", &llt))) {
			  if(!strcasecmp(type, "category")) {
			    ltype = DPS_LIMTYPE_NESTED; fname = DPS_LIMFNAME_CAT;
			  } else
			  if(!strcasecmp(type, "tag")) {
			    ltype = DPS_LIMTYPE_LINEAR_CRC; fname = DPS_LIMFNAME_TAG;
			  } else
			  if(!strcasecmp(type, "time")) {
			    ltype = DPS_LIMTYPE_TIME; fname = DPS_LIMFNAME_TIME;
			  } else
			  if(!strcasecmp(type, "hostname")) {
			    ltype = DPS_LIMTYPE_LINEAR_CRC; fname = DPS_LIMFNAME_HOST;
			  } else
			  if(!strcasecmp(type, "language")) {
			    ltype = DPS_LIMTYPE_LINEAR_CRC; fname = DPS_LIMFNAME_LANG;
			  } else
			  if(!strcasecmp(type, "content")) {
			    ltype = DPS_LIMTYPE_LINEAR_CRC; fname = DPS_LIMFNAME_CTYPE;
			  } else
			  if(!strcasecmp(type, "siteid")) {
			    ltype = DPS_LIMTYPE_LINEAR_INT; fname = DPS_LIMFNAME_SITE;
			  }
			  if((fname != NULL) && strlen(val)) {
			    DpsAddSearchLimit(Agent,ltype,fname,val);
			  }
			}
		}
		tok = dps_strtok_r(NULL, "&", &lt);
	}
	
	DPS_FREE(str);
	DPS_FREE(qs);
	return 0;
}


char * DpsHlConvert(DPS_WIDEWORDLIST *List, const char * src, DPS_CHARSET * lcs, DPS_CHARSET * bcs, int recode) {
	dpsunicode_t	*tok, *lt, *uni;
	int             ctype;
	char		*hpart, *htxt, *zend;
	size_t		len;
	DPS_CONV	lc_uni, uni_bc;
	DPS_CHARSET	*sys_int;
	
	if(!src)return NULL;
	
	if ((len = strlen(src)) == 0) return NULL;
	hpart = (char*)DpsMalloc(len * 14 + 10);
	if (hpart == NULL) return NULL;
	zend = htxt = (char*)DpsMalloc(len * 14 + 10);
	if (htxt == NULL) {
	  DPS_FREE(hpart);
	  return NULL;
	}
	htxt[0]='\0';
	
	sys_int=DpsGetCharSet("sys-int");
	DpsConvInit(&lc_uni, lcs, sys_int, DPS_RECODE_HTML);
	DpsConvInit(&uni_bc, sys_int, bcs, recode);
	
	/* Convert to unicode */
	uni = (dpsunicode_t *)DpsMalloc((len + 10) * sizeof(dpsunicode_t));
	if (uni == NULL) {
	  DPS_FREE(hpart); DPS_FREE(htxt);
	  return NULL;
	}
	DpsConv(&lc_uni,(char*)uni,sizeof(uni[0])*(len+10),src,len+1);
/*	
	if (List != NULL) { 
	  size_t uw; 
	  for(uw = 0; uw < List->nwords; uw++) fprintf(stderr, "w%d:slen:%d|", uw, List->Word[uw].ulen); 
	} else {
	  fprintf(stderr, "List: NULL");
	}
	fprintf(stderr, "\n");
*/

	/* Parse unicode string */
	tok = DpsUniGetSepToken(uni, &lt, &ctype);
	while(tok){
		int found=0;
		size_t slen,flen;
		int euchar;
		size_t uw;

		flen=lt-tok;

		/* Convert token to BrowserCharset */
		euchar=tok[flen];
		tok[flen]=0;
		hpart[0]='\0';

		DpsConv(&uni_bc, hpart, len * 14 + 10, (char*)tok, sizeof(*tok) * flen);
		
		/* Check that it is word to be marked */
		if (List != NULL) {
		  for (uw = 0; uw < List->nwords; uw++) {
			slen = List->Word[uw].ulen;
			if (!DpsUniStrNCaseCmp(tok, List->Word[uw].uword, slen)) {
				found = 1;
				break;
			}
		  }
		}
/*		fprintf(stderr, "tok: %s| flen: %d, found:%d\n", hpart, flen, found);*/
		
		if (found) { *zend = '\2'; zend++; /*strcat(htxt,"\2");*/ }
		/*strcat(htxt,hpart);*/
		strcpy(zend, hpart);
		zend += uni_bc.obytes;
		if (found) { *zend = '\3'; zend++; /*strcat(htxt,"\3");*/ }
		tok[flen]=euchar;

		tok = DpsUniGetSepToken(NULL, &lt, &ctype);
	}
	*zend = '\0';
	DPS_FREE(hpart);
	DPS_FREE(uni);
	/*
	fprintf(stderr,"otxt='%s'\n",src);
	fprintf(stderr,"htxt='%s'\n",htxt);
	*/
	return(htxt);
}

int DpsConvert(DPS_ENV *Conf, DPS_VARLIST *Env_Vars, DPS_RESULT *Res, DPS_CHARSET *lcs, DPS_CHARSET *bcs){
	size_t		i, r, len;
	DPS_CONV	lc_bc, lc_bc_text, bc_bc;
	char            *newval, *newtxt;
	DPS_VAR         *Var;
	
	/* Convert word list */
	DpsConvInit(&lc_bc,lcs,bcs,DPS_RECODE_HTML);
	DpsConvInit(&lc_bc_text, lcs, bcs, 0);
	DpsConvInit(&bc_bc, bcs, bcs, DPS_RECODE_HTML);
	for(i=0;i<Res->WWList.nwords;i++) {
		DPS_WIDEWORD	*W=&Res->WWList.Word[i];

		len = strlen(W->word);
		newval = (char*)DpsMalloc(len * 12 + 1);
		if (newval == NULL) return DPS_ERROR;
		DpsConv(&lc_bc, newval, len * 12 + 1, W->word, len + 1);
		DPS_FREE(W->word);
		W->word = newval;
	}
	
	/* Convert document sections */
	for(i=0;i<Res->num_rows;i++){
		DPS_DOCUMENT	*D=&Res->Doc[i];
		size_t		sec;
		
		for (r = 0; r < 256; r++)
		for (sec = 0; sec < D->Sections.Root[r].nvars; sec++) {
			Var = &D->Sections.Root[r].Var[sec];

			len = strlen(Var->val);
			newtxt = (char*)DpsMalloc(len * 12 + 1);
			if (newtxt == NULL) return DPS_ERROR;

			DpsConv(&lc_bc_text, newtxt, len * 12 + 1, Var->val, len + 1);
			newval = DpsHlConvert(&Res->WWList, Var->val, lcs, bcs, DPS_RECODE_HTML);
			DPS_FREE(Var->val);
			DPS_FREE(Var->txt_val);
			Var->val = newval;
			Var->txt_val = newtxt;
		}
	}

	  /* Convert Env_Vars */
	for (r = 0; r < 256; r++)
	  for (i = 0; i < Env_Vars->Root[r].nvars; i++) {
	        Var = &Env_Vars->Root[r].Var[i];
	        len = strlen(Var->val);
		newtxt = (char*)DpsMalloc(len * 12 + 1);
		newval = (char*)DpsMalloc(len * 12 + 1);
		if (newtxt == NULL || newval == NULL) { DPS_FREE(newtxt); return DPS_ERROR; }

/*		if (db->DBDriver != DPS_DB_SEARCHD)*/  /* FIXME: need unification in charset from different DPS_DB */
		  DpsConv(&lc_bc, newval, len * 12 + 1, Var->val, len + 1);
/*		else 
		  DpsConv(&bc_bc, newval, len * 12 + 1, Var->val, len + 1);*/
		DpsConv(&lc_bc_text, newtxt, len * 12 + 1, Var->val, len + 1);
		DPS_FREE(Var->val);
		DPS_FREE(Var->txt_val);
		Var->val = newval;
		Var->txt_val = newtxt;
	}

	return DPS_OK;
}

char* DpsRemoveHiLightDup(const char *s){
	size_t	len=strlen(s)+1;
	char	*res = (char*)DpsMalloc(len);
	char	*d;

	if (res == NULL) return NULL;
	for(d=res;s[0];s++){
		switch(s[0]){
			case '\2':
			case '\3':
				break;
			default:
				*d++=*s;
		}
	}
	*d='\0';
	return res;
}



int DpsCatToTextBuf(DPS_CATEGORY *C, char *textbuf, size_t len) {
	char	*end;
	size_t  i;
	
	textbuf[0]='\0';

/*	dps_snprintf(textbuf,len,"<CAT");
	end=textbuf+strlen(textbuf);*/
	end = textbuf;
	
	for(i = 0; i < C->ncategories; i++) {
	  dps_snprintf(end, len - strlen(textbuf), "<CAT\tid=\"%d\"\tpath=\"%s\"\tlink=\"%s\"\tname=\"%s\">\r\n",
		   C->Category[i].rec_id, C->Category[i].path, C->Category[i].link, C->Category[i].name );
	  end = end + strlen(end);
	}
/*	strcpy(end,">");*/
	return DPS_OK;
}

int DpsCatFromTextBuf(DPS_CATEGORY *C, char *textbuf) {
	const char	*htok, *last;
	DPS_HTMLTOK	tag;
	size_t		i, c;
	
	if (textbuf == NULL) return DPS_OK;
	DpsHTMLTOKInit(&tag);
	
	htok=DpsHTMLToken(textbuf,&last,&tag);
	
	if(!htok || tag.type != DPS_HTML_TAG)
	  return DPS_OK;

	C->Category = (DPS_CATITEM*)DpsRealloc(C->Category, sizeof(DPS_CATITEM) * ((c = C->ncategories) + 1));
	if (C->Category == NULL) {
	  C->ncategories = 0;
	  return DPS_ERROR;
	}
	bzero((void*)&C->Category[c], sizeof(DPS_CATITEM));
	
	for(i = 1; i < tag.ntoks; i++){
		size_t	nlen = tag.toks[i].nlen;
		size_t	vlen = tag.toks[i].vlen;
		char	*name = DpsStrndup(tag.toks[i].name, nlen);
		char	*data = DpsStrndup(tag.toks[i].val, vlen);

		if (!strcmp(name, "id")) {
		  C->Category[c].rec_id = atoi(data);
		} else if (!strcmp(name, "path")) {
		  strncpy(C->Category[c].path, data, 128);
		} else if (!strcmp(name, "link")) {
		  strncpy(C->Category[c].link, data, 128);
		} else if (!strcmp(name, "name")) {
		  strncpy(C->Category[c].name, data, 128);
		}

		DPS_FREE(name);
		DPS_FREE(data);
	}

	C->ncategories++;
	return DPS_OK;
}

dpsunicode_t *DpsUniSegment(DPS_AGENT *Indexer, dpsunicode_t *ustr, const char *lang) {
	DPS_CHARSET     *tis_cs;
	DPS_CONV        uni_tis, tis_uni;
#ifdef CHASEN
	DPS_CHARSET     *eucjp_cs;
	DPS_CONV        uni_eucjp, eucjp_uni;
	char            *eucstr, *eucstr_seg;
#endif
#ifdef MECAB
	DPS_CHARSET     *sjis_cs;
	DPS_CONV        uni_sjis, sjis_uni;
	char            *sjisstr, *sjisstr_seg;
#endif
#if defined(CHASEN) || defined(MECAB)
	size_t          reslen;
#endif
	DPS_CHARSET	*sys_int;
	size_t          dstlen = DpsUniLen(ustr);

	if (dstlen < 2) return ustr;

	sys_int = DpsGetCharSet("sys-int");

	tis_cs = DpsGetCharSet("tis-620");
	DpsConvInit(&tis_uni, tis_cs, sys_int, DPS_RECODE_HTML);
	DpsConvInit(&uni_tis, sys_int, tis_cs, DPS_RECODE_HTML);
	
#ifdef CHASEN
	eucjp_cs = DpsGetCharSet("euc-jp");
	if (!eucjp_cs) eucjp_cs = DpsGetCharSet("sys-int");
	DpsConvInit(&uni_eucjp, sys_int, eucjp_cs, DPS_RECODE_HTML);
	DpsConvInit(&eucjp_uni, eucjp_cs, sys_int, DPS_RECODE_HTML);
#endif
#ifdef MECAB
	sjis_cs = DpsGetCharSet("sjis");
	if (!sjis_cs) sjis_cs = DpsGetCharSet("sys-int");
	DpsConvInit(&uni_sjis, sys_int, sjis_cs, DPS_RECODE_HTML);
	DpsConvInit(&sjis_uni, sjis_cs, sys_int, DPS_RECODE_HTML);
#endif


#ifdef CHASEN

	if (lang == NULL || !strncasecmp(lang, "ja", 2)) {
	  eucstr = (char*)DpsMalloc(14 * dstlen + 1);
	  if (eucstr == NULL) return NULL;
	  DpsConv(&uni_eucjp, eucstr, 14 * dstlen + 1, (char*)ustr, (dstlen + 1) * sizeof(dpsunicode_t));

	  DPS_GETLOCK(Indexer, DPS_LOCK_SEGMENTER);
	  eucstr_seg = chasen_sparse_tostr(eucstr);
	  DPS_RELEASELOCK(Indexer, DPS_LOCK_SEGMENTER);

	  reslen = strlen(eucstr_seg) + 1;
	  ustr = (dpsunicode_t*)DpsRealloc(ustr, reslen * sizeof(dpsunicode_t));
	  if (ustr == NULL) {
	    DPS_FREE(eucstr);
	    return NULL;
	  }
	  DpsConv(&eucjp_uni, (char*)ustr, reslen * sizeof(dpsunicode_t), eucstr_seg, reslen);
	  DPS_FREE(eucstr);
	  dstlen = DpsUniLen(ustr);
	}
#endif
#ifdef MECAB
	if (lang == NULL || !strncasecmp(lang, "ja", 2)) {
	  sjisstr = (char*)DpsMalloc(12 * dstlen + 1);
	  if (sjisstr == NULL) return NULL;
	  DpsConv(&uni_sjis, sjisstr, 12 * dstlen + 1, (char*)ustr, sizeof(*ustr) * (dstlen + 1));

	  DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
#ifdef HAVE_PTHREADS
	  mecab_lock(Indexer->Conf->mecab);
#endif
	  sjisstr_seg = mecab_sparse_tostr(Indexer->Conf->mecab, sjisstr);
#ifdef HAVE_PTHREADS
	  mecab_unlock(Indexer->Conf->mecab);
#endif
	  DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);

	  reslen = strlen(sjisstr_seg) + 1;
	  ustr = (dpsunicode_t*)DpsRealloc(ustr, reslen * sizeof(dpsunicode_t));
	  if (ustr == NULL) {
	    DPS_FREE(sjisstr); return NULL;
	  }
	  DpsConv(&sjis_uni, (char*)ustr, reslen * sizeof(dpsunicode_t), sjisstr_seg, reslen);
	  DPS_FREE(sjisstr);
	  dstlen = DpsUniLen(ustr);
	}
#endif
#if defined(CHASEN) || defined(MECAB)
/*	else*/
#endif
	  if (lang == NULL || !strncasecmp(lang, "zh", 2)) {
	    dpsunicode_t *seg_ustr;
	    DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
	    seg_ustr = DpsSegmentByFreq(&Indexer->Conf->Chi, ustr);
	    DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);
	    if (seg_ustr != NULL) {
	      DPS_FREE(ustr);
	      ustr = seg_ustr;
	    }
	    dstlen = DpsUniLen(ustr);
	  }
	
#if defined(CHASEN) || defined(MECAB)
/*	  else*/
#endif
	    if (lang == NULL || !strncasecmp(lang, "th", 2)) {
	      dpsunicode_t *seg_ustr;
	      DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
	      seg_ustr = DpsSegmentByFreq(&Indexer->Conf->Thai, ustr);
	      DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);
	      if (seg_ustr != NULL) {
		DPS_FREE(ustr);
		ustr = seg_ustr;
	      }
	      dstlen = DpsUniLen(ustr);
	    }

#if defined(CHASEN) || defined(MECAB)
/*	  else*/
#endif
	    if (lang == NULL || !strncasecmp(lang, "ko", 2)) {
	      dpsunicode_t *seg_ustr;
	      DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
	      seg_ustr = DpsSegmentByFreq(&Indexer->Conf->Korean, ustr);
	      DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);
	      if (seg_ustr != NULL) {
		DPS_FREE(ustr);
		ustr = seg_ustr;
	      }
	      dstlen = DpsUniLen(ustr);
	    }
	return ustr;
}


char * DpsBuildPageURL(DPS_VARLIST * vars, char **dst) {
	size_t i, r, nargs = 0, dstlen = 0;
	char * end;

	for (r = 0; r < 256; r++)
	for(i = 0; i < vars->Root[r].nvars; i++) {
	  dstlen += 7 + strlen(vars->Root[r].Var[i].name) + strlen(vars->Root[r].Var[i].val);
	}
	*dst = (char*)DpsRealloc(*dst, dstlen);
	if (*dst == NULL) return NULL;
	end = *dst;

	for (r = 0; r < 256; r++)
	for (i = 0; i < vars->Root[r].nvars; i++) {
		strcpy(end, nargs ? "&amp;" : "?");
		end += nargs ? 5 : 1;
		strcpy(end, vars->Root[r].Var[i].name);
			
		end = end + strlen(end);
		strcpy(end, "="); end++;
		strcpy(end, vars->Root[r].Var[i].val);
			
		end = end + strlen(end);
		nargs++;
	}
	return NULL;
}

void DpsParseQStringUnescaped(DPS_VARLIST *vars, const char *qstring){
	char *tok, *lt; 
	char *qs = (char*)DpsStrdup(qstring);

	if (qs != NULL) {
	  DpsSGMLUnescape(qs);
	  tok = dps_strtok_r(qs, "&", &lt);
	  while(tok){
	        char *arg=strchr(tok,'=');
		if(arg)*arg++='\0';
		DpsVarListAddStr(vars,tok,arg?arg:"");
		tok = dps_strtok_r(NULL, "&", &lt);
	  }
	  DpsFree(qs);
	}
}
